'' *****************************
'' GPS routines
''  (c) 2007 Perry James Mole
''  pjm@ridge-communications.ca
''
''  Tim Moore May 2008
''    Support app COG
''    Fix some bugs
''    Merge in FV-E8 support (WAAS, switch to 38400 baud, etc)
''    Use pcFullduplexserial4
''  May 2009
''    Add conversion of parts to numbers
''    Move port configuration into here
''  May 2010
''    Add conversion to Qs15_16 for altitude and speed
''    Add PBM-248 configuration
''  June 2010
''    Add conversion to Qs15_16 for latitude and longtude
'' *****************************

' PVH Comment - Excellent small GPS reader routines.
' $GPRMC  Recommended minimum data                 ie: $GPRMC,081836,A,3751.6565,S,14507.3654,E,000.0,360.0,130998,011.3,E*62
' $GPGGA  GPS Fix Data                             ie: $GPGGA,170834,4124.8963,N,08151.6838,W,1,05,1.5,280.2,M,-34.0,M,,,*75
' $PGRMZ  eTrex proprietary barametric altitude ft ie: $PGRMZ,453,f,2*18

CON
  CR = 13                                             ' ASCII <CR>
  LF = 10                                             ' ASCII <LF>

' NMEASTate
  WAITFORSTARTLINE = 0
  WAITFORENDLINE   = 1

VAR  
#ifdef USEGPSCOG
  long gps_stack[10]                                    '
  long cog
#endif
  long GPRMCa[20],GPGGAa[20]',PGRMZa[20]   
  long cptr
  long NMEAState
  long port

  byte GPRMCb[68],GPGGAb[80]',PGRMZb[40]
  byte gps_buff[80]',cksum
  byte buf[20], minpart[3]

OBJ
  uarts         : "pcfullduplexserial4FC"
  format        : "obex_Format"       
  trig          : "imumathq"                            '0 Cog

#ifdef USEGPSCPG
PUB Start(portp, rxpin, txpin, type) : okay
'' Starts uart object (at baud specified) in a cog
'' -- returns false if no cog available
  okay := Init(portp, rxpin, txpin, type)

  return cog := cognew(readNEMA,@gps_stack) + 1 
#endif

PUB Init(portp, rxpin, txpin, type) : okay
''Init GPS serial without a COG to read from port
''If app calls this directly rather than Start then app needs to call GetNMEALine repeatly
  port := portp
  cptr := 0
  NMEAState := WAITFORSTARTLINE
  uarts.ReInit
  uarts.AddPort(port,rxpin,txpin,UARTS#PINNOTUSED,UARTS#PINNOTUSED,UARTS#DEFAULTTHRESHOLD, {
}   UARTS#NOMODE,UARTS#BAUD4800)                        'Add gps port
  uarts.Start
  if type <> 0 AND txpin <> UARTS#PINNOTUSED            'if no tx pin GPS must be 4800baud NMEA
    configbaud(type)                                    'Change GPS baudrate
    uarts.ReInit
    uarts.AddPort(port,rxpin,txpin,UARTS#PINNOTUSED,UARTS#PINNOTUSED,UARTS#DEFAULTTHRESHOLD, {
}   UARTS#NOMODE,UARTS#BAUD38400)                       'Add gps port
    uarts.Start
    configgps(type)                                     'Config rest of gps settings

PUB configbaud(type)
  waitcnt(clkfreq + cnt)                                'Delay while serial initializes
  case type
    1:
      uarts.str(port,string("$PMTK251,38400*27",CR,LF)) 'configure GPS to use 38400 baud
    2:
      uarts.str(port,string("@CB 38400",CR,LF))         'configure GPS to use 38400 baud
  uarts.txflush(port)                                   'Wait until all chars txed
  waitcnt(clkfreq + cnt)                                'delay for last char to be sent

'split into separate config since some of these will not work until baud rate has been changed
PUB configgps(type)
  'configure GPS to send only the strings i want (GGA and RMC)
  case type
    1:
      uarts.str(port,string("$PMTK314,0,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*28",CR,LF))
      uarts.str(port,string("$PMTK313,1*2E",CR,LF))     'configure GPS to use WAAS
      uarts.str(port,string("$PMTK301,2*2E",CR,LF))     'configure GPS to use WAAS
      uarts.str(port,string("$PMTK220,200*2C",CR,LF))   'configure GPS to update at 5Hz
    2:
      uarts.str(port,string("@WLK ON",CR,LF))           'first wlk is ignored
      uarts.str(port,string("@WLK ON",CR,LF))           'need to send WLK twice
      uarts.str(port,string("@WA ON 0",CR,LF))          'WAAS on
      uarts.str(port,string("@NC 10001000",CR,LF))      'only GGA and RMC output

#ifdef USEGPSCPG
PUB readNEMA
' if start is called this processes reads from GPS on a COG
  repeat
    GetNMEALine
#endif

PUB GetNMEALine | Rx
'Needs to be called often enough that uart rx buffer doesn't overflow
'May never return if serial receive rate is faster than processing rate
  result := false

  'process all characters in uart receive buffer
  repeat while (Rx := uarts.rxcheck(port)) <> -1
    'uarts.tx(0,Rx)
    'uarts.hex(0,Rx,2)
    'uarts.tx(0," ")
    CASE NMEAState
      WAITFORSTARTLINE:
        if Rx == "$"                                    'wait for the $ to insure we are starting with
          NMEAState := WAITFORENDLINE                   'a complete NMEA sentence
          bytefill(@gps_buff,0,80)                      'empty line buffer
          cptr := 0
      WAITFORENDLINE:
        if cptr < 80
          CASE Rx                                       'continue to collect data until the end of the NMEA sentence
            "," :
              gps_buff[cptr++] := 0                     'If "," replace the character with 0
            CR:
              ProcessNMEALine                           'Process the just received line
              NMEAState := WAITFORSTARTLINE             'Start looking for start of NMEA line
              result := true
              cptr := 0
            OTHER:
              gps_buff[cptr++] := Rx                    'else save the character        
        else
'          uarts.strln(0,string("GPS BUFFER OVERFLOW")) 'buffer overflowed, get a new line
          NMEAState := WAITFORSTARTLINE                 'Start looking for start of NMEA line
          cptr := 0

PUB ProcessNMEALine

  if gps_buff[2] == "G"
    if gps_buff[3] == "G"            
      if gps_buff[4] == "A"            
        copy_buffer(@GPGGAb, @GPGGAa)

  if gps_buff[2] == "R"             
    if gps_buff[3] == "M"            
      if gps_buff[4] == "C"           
        copy_buffer(@GPRMCb, @GPRMCa)

{  ' Garmin specific                   
  if gps_buff[0] == "P"
    if gps_buff[1] == "G"  
      if gps_buff[2] == "R"
        if gps_buff[3] == "M"  
          if gps_buff[4] == "Z"
            copy_buffer(@PGRMZb, @PGRMZa)
}

pub copy_buffer (buffer,args) | ptr, arg
  bytemove(buffer,@gps_buff,cptr)   'copy received data to buffer
  ptr := buffer
  arg := 0
  repeat 79                                             'build array of pointers
    if byte[ptr] == 0               'to each
      if byte[ptr+1] == 0           'record
        long[args+4*arg] := 0       'in
      else                          'the
        long[args+4*arg] := ptr+1   'data buffer
      if arg < 20
        arg++
    ptr++

pub altitude
' now we just need to return the pointer to the desired record
   return GPGGAa[8]

pub valid
   return GPRMCa[1]

pub IsValid
   return byte[GPRMCa[1]] == "A"

pub speed
   if byte[GPRMCa[6]] => "0" AND byte[GPRMCa[6]] <= "0" 
     return GPRMCa[6]
   else
     return @empty 

pub heading
   if byte[GPRMCa[7]] => "0" AND byte[GPRMCa[7]] <= "0" 
     return GPRMCa[7]
   else
     return @empty 

pub date
   return GPRMCa[8]

pub GPSaltitude
   return GPGGAa[8]

pub time
   return GPGGAa[0]

pub latitude
   return GPGGAa[1]

pub N_S
   return GPGGAa[2]

pub longitude
   return GPGGAa[3]

pub E_W
   return GPGGAa[4]

pub satellites
   return GPGGAa[6]

pub hdop
   return GPGGAa[7]

pub vdop
   return GPGGAa[14]

pub noGPSData
   GPRMCa.long[1] := string("X")

PUB crc8bitwise(pBuf, buflen, poly, initial): crc | i,k
'pBuf is the string to be CRCs
'buflen is thelength of th ebuffer to be CRCd
'poly is the ?
'initial is the inital value of CRC
 
  crc := initial
  repeat i from 0 to buflen - 1
    crc ^= (BYTE[pBuf][i] >< 8)
    repeat k from 0 to 7
      if crc & $80
        crc := (crc << 1) ^ poly
      else
        crc <<= 1
  crc &= $ff
  crc ><= 8

pri to_hex(key)

  case key
   "0": return 0
   "1": return 1
   "2": return 2
   "3": return 3 
   "4": return 4
   "5": return 5
   "6": return 6
   "7": return 7
   "8": return 8
   "9": return 9
   "A": return 10
   "B": return 11
   "C": return 12
   "D": return 13
   "E": return 14
   "F": return 15

pub headingDeg | h
  h := format.atoi(heading)*10
  return h

pub QheadingDeg | intp, fracp
'' heading in degrees, , Qs15_16 format
  result := Split(heading, @intp, @fracp)
  result := (intP << 16) + trig.Qdiv(fracP<<16, result)

pub QspeedInMph | intp, fracp
'' Speed over the ground in MPH, Qs15_16 format
  result := Split(GPSaltitude, @intp, @fracp)
  result := trig.Qmul((intP << 16) + trig.Qdiv(fracP<<16, result), constant((trunc((11.52))<<16)|round(((11.52)-float(trunc((11.52))))*65536.0)))

pub speedInMph | mph
'' Speed over the ground in MPH * 10 (returns a number)
   mph := (format.atoi(speed)*1152)/100
   return mph 

pub altitudeFeet | feet
'' Altitude, ft, above mean sea level * 10
  feet := (Format.atoi(GPSaltitude)* 32808)/1000
  return feet

pub QaltitudeM | intp, fracp
'' Altitude, m, above mean sea level, Qs15_16 format
  result := Split(GPSaltitude, @intp, @fracp)
  result := (intP << 16) + trig.QDiv(fracP<<16, result)

pub QLatitude | dd, mm, mmmm
  result := SplitL(latitude, @dd, @mm, @mmmm)
  result := (dd << 16) + trig.QDiv((mm<<16)+trig.QDiv(mmmm<<16, result), 60)

pub QLongitude | dd, mm, mmmm
  result := SplitL(longitude, @dd, @mm, @mmmm)
  result := (dd << 16) + trig.QDiv((mm<<16)+trig.QDiv(mmmm<<16, result), 60)

pub timehh  | hh
  hh := to_hex(byte[time][0]) *10   'uses existing hex lookup as a decimal lookup
  hh += to_hex(byte[time][1])
  return hh

pub timemm  | mm
  mm := to_hex(byte[time][2]) *10   'uses existing hex lookup as a decimal lookup
  mm += to_hex(byte[time][3])
  return mm

pub timess  | ss
  ss := to_hex(byte[time][4]) *10   'uses existing hex lookup as a decimal lookup
  ss += to_hex(byte[time][5])
  return ss

pub timems  | ms
  ms := to_hex(byte[time][7]) *100  'uses existing hex lookup as a decimal lookup
  ms += to_hex(byte[time][8]) *10
  ms += to_hex(byte[time][9])
  return ms

pub dateDay | dd
  dd := to_hex(byte[date][0]) *10   'uses existing hex lookup as a decimal lookup
  dd += to_hex(byte[date][1])
  return dd

pub dateMonth  | mm
  mm := to_hex(byte[date][2]) *10   'uses existing hex lookup as a decimal lookup
  mm += to_hex(byte[date][3])
  return mm

pub dateYear  | yy
  yy := to_hex(byte[date][4]) *10   'uses existing hex lookup as a decimal lookup
  yy += to_hex(byte[date][5])
  return yy + 2000

pub satelliteCount | n
  '' Number of satellites being tracked (returns a number)
  if(byte[satellites][1] == 0)
    n := to_hex(byte[satellites][0])       'uses existing hex lookup as a decimal lookup
  else
    n := to_hex(byte[satellites][0]) *10   'uses existing hex lookup as a decimal lookup
    n += to_hex(byte[satellites][1])

  return n

pub SplitL(sPtr, ddptr, mmptr, mmmmptr) | i, c, fractPartPtr
    bytefill(@buf,0,20)                                 'clear buffer
    bytemove(@buf,sPtr,strsize(sPtr))                   'Make a copy to mess with

    i := 0

    repeat                                              'Locate the '.' (decimal point)  
      c := buf[ i++ ]
      if i == 20
        return                                          'in case invalid string
    until c == "."

    buf[ --i ] := 0                                     'split into 2 strings by replacing '.' with null

    fractPartPtr := @buf[ i+1 ]                         'Fractional part is what was after the '.'
    i -= 2                                              'back up 2 bytes to get the minutes
    bytemove(@minPart[ 0 ],@buf[ i ],2)                 'copy the mm part
    minPart[ 2 ] := 0                                   'null terminate the mm part
    buf[i] := 0                                         'null terminate to remove the mm part

    byte[ddptr]   := Format.atoi(@buf)                  'Convert to number
    byte[mmptr]   := Format.atoi(@minPart)
    long[mmmmptr] := Format.atoi(fractPartPtr)

    result := Len10s(fractPartPtr)

pub Split(sPtr, ddptr, fptr) | i, c, fractPartPtr
    bytefill(@buf,0,20)                                 'clear buffer
    bytemove(@buf,sPtr,strsize(sPtr))                   'Make a copy to mess with

    i := 0

    repeat                                              'Locate the '.' (decimal point)  
      c := buf[ i++ ]
      if i == 20
        return                                          'in case invalid string
    until c == "."

    buf[ --i ] := 0                                     'split into 2 strings by replacing '.' with null

    fractPartPtr := @buf[ i+1 ]                         'Fractional part is what was after the '.'

    byte[ddptr]   := Format.atoi(@buf)                  'Convert to number
    long[fptr] := Format.atoi(fractPartPtr)

    result := Len10s(fractPartPtr)

pub LatitudeParts(ddptr,mmptr, mmmmptr)
  return SplitL(latitude, ddptr, mmptr, mmmmptr)

pub LongitudeParts(ddptr,mmptr, mmmmptr)
  return SplitL(longitude, ddptr, mmptr, mmmmptr)

PRI Len10s(PPtr)

  result := 1
  repeat byte[PPtr++] <> 0
    result *= 10

DAT
empty   byte  $20, $00
{{
}}
